"""
Stats on transcription data
"""
from clldutils.markup import Table

from pyclts.models import is_valid_sound
from pyclts.cli_util import add_format


def register(parser):
    add_format(parser, default='simple')


def run(args):
    table = Table('id', 'valid', 'total', 'percent')
    bipa = args.repos.bipa
    for td in args.repos.iter_transcriptiondata():
        ln = [1 if is_valid_sound(bipa[name], bipa) else 0 for name in td.names]
        table.append([td.id, sum(ln), len(ln), sum(ln) / len(ln)])
    table.append([
        len(table) - 1,
        '',
        '',
        0 if not len(table) - 1 else sum([line[-1] for line in table[1:]]) / (len(table) - 1)])
    print(table.render(tablefmt=args.format))



"""
Display basic info about sounds
"""
from pyclts.cli_util import add_format, Table


def register(parser):
    add_format(parser)
    parser.add_argument(
        'sounds',
        metavar='SOUNDS',
        nargs='+',
        help='sounds to display info for')


def run(args):
    tts = args.repos.transcriptionsystem(args.system)
    with Table(args, args.system.upper(), 'SOURCE', 'GENERATED', 'ALIAS', 'NAME') as data:
        for sound in args.sounds:
            sound = tts.get(sound if isinstance(sound, str) else sound.decode('utf8'))
            if sound.type != 'unknownsound':
                data.append([
                    str(sound),
                    sound.source or ' ',
                    '1' if sound.generated else ' ',
                    sound.grapheme if sound.alias else ' ',
                    sound.name,
                ])
            else:
                data.append(['?', sound.source, '?', '?', '?'])



"""

"""
import collections

from csvw.dsv import reader

from pyclts.cli_util import add_format, Table


def register(parser):
    add_format(parser)


def run(args):
    def read(fname):
        return reader(args.repos.path('data', fname), delimiter='\t', dicts=True)

    sounds = {row['NAME']: row for row in read('sounds.tsv')}
    graphs = {'{GRAPHEME}-{NAME}-{DATASET}'.format(**row): row for row in read('graphemes.tsv')}

    graphdict = collections.defaultdict(list)
    for id_, row in graphs.items():
        graphdict[row['GRAPHEME']].append(row['DATASET'])

    with Table(args, 'DATA', 'STATS', 'PERC') as text:
        text.append(
            ['Unique graphemes', len(set(row['GRAPHEME'] for row in graphs.values())), ''])
        text.append(['different sounds', len(sounds), ''])
        text.append(
            ['singletons', len([g for g in graphdict if len(set(graphdict[g])) == 1]), ''])
        text.append(
            ['multiples', len([g for g in graphdict if len(set(graphdict[g])) > 1]), ''])
        total = len(sounds)
        for type_, count in collections.Counter([s['TYPE'] for s in sounds.values()]).most_common():
            text.append([type_ + 's', count, count / total])



"""
Transcription System module for consistent IPA handling.
========================================

"""
import re

from csvw import TableGroup
from clldutils import jsonlib
import attr

from pyclts.util import nfd, norm, EMPTY, itertable, TranscriptionBase
from pyclts.models import *  # noqa: F403


class TranscriptionSystem(TranscriptionBase):
    """
    A transcription System."""
    __type__ = 'ts'

    def __init__(self, path, metadata, features):
        """
        :param system: The name of a transcription system or a directory containing one.
        """
        super().__init__(path, None)
        if not (self.path.exists() and self.path.is_dir()):
            raise ValueError('unknown system: {0}'.format(self.path))

        self.system = TableGroup.from_file(metadata)
        self.system._fname = path / 'metadata.json'

        self.features = {'consonant': {}, 'vowel': {}, 'tone': {}}
        # dictionary for feature values, checks when writing elements from
        # write_order to make sure no output is doubled
        self._feature_values = {}

        # load the general features
        features = jsonlib.load(features)

        self.diacritics = dict(
            consonant={}, vowel={}, click={}, diphthong={}, tone={}, cluster={})
        for dia in itertable(self.system.tabledict['diacritics.tsv']):
            if not dia['alias'] and not dia['typography']:
                self.features[dia['type']][dia['value']] = dia['grapheme']
            # assign feature values to the dictionary
            self._feature_values[dia['value']] = dia['feature']
            self.diacritics[dia['type']][dia['grapheme']] = dia['value']

        self.sound_classes = {}
        self.columns = {}  # the basic column structure, to allow for rendering
        self.sounds = {}  # Sounds by grapheme
        self._covered = {}
        # check for unresolved aliased sounds
        aliases = []
        for cls in [Consonant, Vowel, Tone, Marker]:  # noqa: F405
            type_ = cls.__name__.lower()
            self.sound_classes[type_] = cls
            # store information on column structure to allow for rendering of a
            # sound in this form, which will make it easier to insert it when
            # finding generated sounds
            self.columns[type_] = [
                c['name'].lower() for c in
                self.system.tabledict['{0}s.tsv'.format(type_)]
                    .asdict()['tableSchema']['columns']]
            for l, item in enumerate(itertable(
                    self.system.tabledict['{0}s.tsv'.format(type_)])):
                if item['grapheme'] in self.sounds:
                    raise ValueError('duplicate grapheme in {0}:{1}: {2}'.format(
                        type_ + 's.tsv', l + 2, item['grapheme']))
                sound = cls(ts=self, **item)
                # make sure this does not take too long
                for key, value in item.items():
                    if key not in {'grapheme', 'note', 'alias'} and \
                            value and value not in self._feature_values:
                        self._feature_values[value] = key
                        if type_ != 'marker' and value not in features[type_][key]:
                            raise ValueError(
                                "Unrecognized features ({0}: {1}, line {2}))".format(
                                    key, value, l + 2))

                self.sounds[item['grapheme']] = sound
                if not sound.alias:
                    if sound.featureset in self.features:
                        raise ValueError('duplicate features in {0}:{1}: {2}'.format(
                            type_ + 's.tsv', l + 2, sound.name))
                    self.features[sound.featureset] = sound
                else:
                    aliases += [(l, sound.type, sound.featureset)]
        # check for consistency of aliases: if an alias has no counterpart, it
        # is orphaned and needs to be deleted or given an accepted non-aliased
        # sound
        if [x for x in aliases if x[2] not in self.features]:  # pragma: no cover
            error = ', '.join(
                str(x[0] + 2) + '/' + str(x[1])
                for x in aliases if x[2] not in self.features)
            raise ValueError(
                'Orphaned aliases in line(s) {0}'.format(error))

        # basic regular expression, used to match the basic sounds in the system.
        self._regex = None
        self._update_regex()

        # normalization data
        self._normalize = {
            norm(r['source']): norm(r['target'])
            for r in itertable(self.system.tabledict['normalize.tsv'])}

    def _update_regex(self):
        self._regex = re.compile('|'.join(
            map(re.escape, sorted(self.sounds, key=lambda x: len(x), reverse=True))))

    def _norm(self, string):
        """Extended normalization: normalize by list of norm-characers, split
        by character "/"."""
        nstring = norm(string)
        if "/" in string:
            s, t = string.split('/')
            nstring = t
        return self.normalize(nstring)

    def normalize(self, string):
        """Normalize the string according to normalization list"""
        return ''.join([self._normalize.get(x, x) for x in nfd(string)])

    def _from_name(self, string):
        """Parse a sound from its name"""
        components = string.split(' ')
        if frozenset(components) in self.features:
            return self.features[frozenset(components)]
        rest, sound_class = components[:-1], components[-1]
        if sound_class in ['diphthong', 'cluster']:
            if string.startswith('from ') and 'to ' in string:
                extension = {'diphthong': 'vowel', 'cluster': 'consonant'}[sound_class]
                string_ = ' '.join(string.split(' ')[1:-1])
                from_, to_ = string_.split(' to ')
                v1, v2 = frozenset(from_.split(' ') + [extension]), frozenset(
                    to_.split(' ') + [extension])
                if v1 in self.features and v2 in self.features:
                    s1, s2 = (self.features[v1], self.features[v2])
                    if sound_class == 'diphthong':
                        return Diphthong.from_sounds(s1 + s2, s1, s2, self)  # noqa: F405
                    else:
                        return Cluster.from_sounds(s1 + s2, s1, s2, self)  # noqa: F405
                else:
                    # try to generate the sounds if they are not there
                    s1, s2 = self._from_name(from_ + ' ' + extension), self._from_name(
                        to_ + ' ' + extension)
                    if not (isinstance(
                        s1, UnknownSound) or isinstance(s2, UnknownSound)):  # noqa: F405
                        if sound_class == 'diphthong':
                            return Diphthong.from_sounds(  # noqa: F405
                                s1 + s2, s1, s2, self)
                        return Cluster.from_sounds(s1 + s2, s1, s2, self)  # noqa: F405
                    raise ValueError('components could not be found in system')
            else:
                raise ValueError('name string is erroneously encoded')

        if sound_class not in self.sound_classes:
            raise ValueError('no sound class specified')

        args = {self._feature_values.get(comp, '?'): comp for comp in rest}
        if '?' in args:
            raise ValueError('string contains unknown features')
        args['grapheme'] = ''
        args['ts'] = self
        sound = self.sound_classes[sound_class](**args)
        if sound.featureset not in self.features:
            sound.generated = True
            return sound
        return self.features[sound.featureset]

    def _parse(self, string):
        """Parse a string and return its features.

        :param string: A one-symbol string in NFD

        Notes
        -----
        Strategy is rather simple: we determine the base part of a string and
        then search left and right of this part for the additional features as
        expressed by the diacritics. Fails if a segment has more than one basic
        part.
        """
        nstring = self._norm(string)

        # check whether sound is in self.sounds
        if nstring in self.sounds:
            sound = self.sounds[nstring]
            sound.normalized = nstring != string
            sound.source = string
            return sound

        match = list(self._regex.finditer(nstring))
        # if the match has length 2, we assume that we have two sounds, so we split
        # the sound and pass it on for separate evaluation (recursive function)
        if len(match) == 2:
            sound1 = self._parse(nstring[:match[1].start()])
            sound2 = self._parse(nstring[match[1].start():])
            # if we have ANY unknown sound, we mark the whole sound as unknown, if
            # we have two known sounds of the same type (vowel or consonant), we
            # either construct a diphthong or a cluster
            if 'unknownsound' not in (sound1.type, sound2.type) and \
                    sound1.type == sound2.type:
                # diphthong creation
                if sound1.type == 'vowel':
                    return Diphthong.from_sounds(  # noqa: F405
                        string, sound1, sound2, self)
                elif sound1.type == 'consonant' and \
                        sound1.manner in ('stop', 'implosive', 'click', 'nasal') and \
                        sound2.manner in ('stop', 'implosive', 'affricate', 'fricative'):
                    return Cluster.from_sounds(  # noqa: F405
                        string, sound1, sound2, self)
            return UnknownSound(grapheme=nstring, source=string, ts=self)  # noqa: F405

        if len(match) != 1:
            # Either no match or more than one; both is considered an error.
            return UnknownSound(grapheme=nstring, source=string, ts=self)  # noqa: F405

        pre, mid, post = nstring.partition(nstring[match[0].start():match[0].end()])
        base_sound = self.sounds[mid]
        if isinstance(base_sound, Marker):  # noqa: F405
            assert pre or post
            return UnknownSound(grapheme=nstring, source=string, ts=self)  # noqa: F405

        # A base sound with diacritics or a custom symbol.
        features = attr.asdict(base_sound)
        features.update(
            source=string,
            generated=True,
            normalized=nstring != string,
            base=base_sound.grapheme)

        # we construct two versions: the "normal" version and the version where
        # we search for aliases and normalize them (as our features system for
        # diacritics may well define aliases
        grapheme, sound = '', ''
        for dia in [p + EMPTY for p in pre]:
            feature = self.diacritics[base_sound.type].get(dia, {})
            if not feature:
                return UnknownSound(  # noqa: F405
                    grapheme=nstring, source=string, ts=self)
            features[self._feature_values[feature]] = feature
            # we add the unaliased version to the grapheme
            grapheme += dia[0]
            # we add the corrected version (if this is needed) to the sound
            sound += self.features[base_sound.type][feature][0]
        # add the base sound
        grapheme += base_sound.grapheme
        sound += base_sound.s
        for dia in [EMPTY + p for p in post]:
            feature = self.diacritics[base_sound.type].get(dia, {})
            # we are strict: if we don't know the feature, it's an unknown
            # sound
            if not feature:
                return UnknownSound(  # noqa: F405
                    grapheme=nstring, source=string, ts=self)
            features[self._feature_values[feature]] = feature
            grapheme += dia[1]
            sound += self.features[base_sound.type][feature][1]

        features['grapheme'] = sound
        new_sound = self.sound_classes[base_sound.type](**features)
        # check whether grapheme differs from re-generated sound
        if str(new_sound) != sound:
            new_sound.alias = True
        if grapheme != sound:
            new_sound.alias = True
            new_sound.grapheme = grapheme
        return new_sound

    def resolve_sound(self, string):
        if isinstance(string, Sound):  # noqa: F405
            return self.features[string.featureset]
        elif isinstance(string, Symbol):  # noqa: F405
            return string
        if set(string.split(' ')).intersection(
                list(self.sound_classes) + ['diphthong', 'cluster']):
            return self._from_name(string)
        string = nfd(string)
        return self._parse(string)

    def __contains__(self, item):
        if isinstance(item, Sound):  # noqa: F405
            return item.featureset in self.features
        return item in self.sounds

    def __iter__(self):
        return iter(self.sounds)



from pyclts.transcriptionsystem import Symbol
from pyclts.util import read_data, TranscriptionBase

SOUNDCLASS_SYSTEMS = ['sca', 'cv', 'art', 'dolgo', 'asjp', 'color']


class SoundClasses(TranscriptionBase):
    """
    Class for handling sound class models.
    """
    __type__ = 'sc'

    def __init__(self, path, system, id_):
        assert id_ in SOUNDCLASS_SYSTEMS
        super().__init__(path, system)
        self._id = id_
        _, data, self.sounds, self.names = read_data(self.path, self._id)
        self.data = {}
        self.classes = set()
        for k, v in data.items():
            self.data[k] = v[0]
            self.classes.add(v[0]['grapheme'])

    @property
    def id(self):
        return self._id

    def resolve_sound(self, sound):
        """Function tries to identify a sound in the data.

        Notes
        -----
        The function tries to resolve sounds to take a sound with less complex
        features in order to yield the next approximate sound class, if the
        transcription data are sound classes.
        """
        sound = sound if isinstance(sound, Symbol) else self.system[sound]
        if sound.name in self.data:
            return self.data[sound.name]['grapheme']
        if not sound.type == 'unknownsound':
            if sound.type in ['diphthong', 'cluster']:
                return self.resolve_sound(sound.from_sound)
            name = [
                s for s in sound.name.split(' ') if
                self.system._feature_values.get(s, '') not in
                ['laminality', 'ejection', 'tone']]
            while len(name) >= 4:
                sound = self.system.get(' '.join(name))
                if sound and sound.name in self.data:
                    return self.resolve_sound(sound)
                name.pop(0)
        raise KeyError(":sc:resolve_sound: No sound could be found.")



"""
Create the CLTS javascript app.
"""
import collections
import json


def run(args, test=False):
    tts = args.repos.bipa

    def sound_to_dict(snd):
        res = collections.OrderedDict([('name', snd.name), ('bipa', snd.s), ('type', snd.type)])
        for f in snd._name_order:
            res[f] = getattr(snd, f)
        return res

    # retrieve all sounds in the datasets
    all_sounds = collections.OrderedDict()
    for td in args.repos.iter_transcriptiondata():
        for sound in td.data:
            if ' ' in sound:
                snd = tts[sound]
                glyph = snd.s
                assert '<?>' not in snd.s
                if snd.s not in all_sounds:
                    all_sounds[glyph] = sound_to_dict(snd)
                for item in td.data[sound]:
                    if item['grapheme'] not in all_sounds:
                        all_sounds[item['grapheme']] = all_sounds[glyph]

                all_sounds[glyph][td.id] = td.data[sound]
        if test:
            break

    # add sounds from transcription system
    for sound in tts:
        if sound not in all_sounds:
            snd = tts[sound]
            if snd.type != 'marker':
                if snd.s in all_sounds:
                    all_sounds[sound] = all_sounds[snd.s]
                else:
                    all_sounds[sound] = sound_to_dict(snd)

    args.log.info('{0} unique graphemes loaded'.format(len(all_sounds)))

    for i, sc in enumerate(args.repos.iter_soundclass()):
        for sound in all_sounds:
            try:
                all_sounds[sound][sc.id] = [dict(grapheme=sc[sound])]
            except KeyError:  # pragma: no cover
                pass
            if i == 0:
                if hasattr(sound, 's'):
                    all_sounds[sound]['bipa'] = tts[sound].s
        if test:
            break

    datafile = args.repos.repos / 'app' / 'data.js'
    with datafile.open('w', encoding='utf8') as handler:
        handler.write('var BIPA = ' + json.dumps(all_sounds, indent=2) + ';\n')
        handler.write('var normalize = ' + json.dumps(tts._normalize) + ';\n')
    args.log.info('{0} written'.format(datafile))



import unicodedata

import attr
from clldutils.misc import nfilter, lazyproperty

from pyclts.util import norm

__all__ = [
    'is_valid_sound',
    'Symbol',
    'Sound',
    'Consonant',
    'Vowel',
    'Tone',
    'Marker',
    'Diphthong',
    'Cluster',
    'UnknownSound']
EXCLUDE_FEATURES = [
    'apical',
    'laminal',
    'ejective',
    'with_falling_tone',
    'with_extra-low_tone',
    'with_extra-high_tone',
    'with_falling_tone',
    'with_low_tone',
    'with_global_fall',
    'with_global_rise',
    'with_high_tone',
    'with_mid_tone',
    'with_rising_tone',
    'with_upstep'
]


def is_valid_sound(sound, ts):
    """Check the consistency of a given transcription system conversino"""
    if isinstance(sound, (Marker, UnknownSound)):
        return False
    s1 = ts[sound.name]
    s2 = ts[sound.s]
    return s1.name == s2.name and s1.s == s2.s


@attr.s(cmp=False)
class Symbol(object):
    ts = attr.ib()
    grapheme = attr.ib()
    source = attr.ib(default=None)
    generated = attr.ib(default=False, validator=attr.validators.instance_of(bool))
    note = attr.ib(default=None)

    @lazyproperty
    def type(self):
        return self.__class__.__name__.lower()

    def __str__(self):
        return self.grapheme

    def __eq__(self, other):
        """
        In the absence of features, we consider symbols equal, if they belong to the same
        system and are represented by the same grapheme.
        """
        return self.ts.id == other.ts.id and self.grapheme == other.grapheme

    @property
    def name(self):
        return None

    @property
    def uname(self):
        "Return unicode name(s) for a character set."
        try:
            return ' / '.join(unicodedata.name(ss) for ss in str(self))
        except TypeError:
            return '-'
        except ValueError:
            return '?'

    @property
    def codepoints(self):
        "Return unicode codepoint(s) for a grapheme."
        return ' '.join('U+' + ('000' + hex(ord(x))[2:])[-4:] for x in str(self))


@attr.s(cmp=False)
class UnknownSound(Symbol):
    pass


@attr.s(cmp=False, repr=False)
class Sound(Symbol):
    """
    Sound object stores basic features of the individual sound objects.
    """
    base = attr.ib(default=None)
    alias = attr.ib(default=None)
    normalized = attr.ib(default=None)
    unknown = attr.ib(default=None)
    stress = attr.ib(default=None)

    _name_order = []
    _write_order = dict(pre=[], post=[])

    def __eq__(self, other):
        if isinstance(other, Sound):
            return self.name == other.name
        return False

    def __repr__(self):
        return '<{0}.{1}: {2}>'.format(
            self.__module__, self.__class__.__name__, self.name)

    def __add__(self, other):
        return str(self) + str(other)

    def __hash__(self):
        return hash(self.name)

    @property
    def s(self):
        return str(self)

    def _features(self):
        return nfilter(getattr(self, p, None) for p in self._name_order)

    @property
    def featuredict(self):
        return {f: getattr(self, f, None) for f in self._name_order}

    @property
    def featureset(self):
        return frozenset(self._features() + [self.type])

    def similarity(self, other):
        f1, f2 = self.featureset, other.featureset
        return len(f1.intersection(f2)) / len(f1.union(f2))

    def __str__(self):
        """
        Return the reference representation of the sound.

        Note
        ----
        We first try to return the non-alias value in our data. If this fails,
        we create the sound based on it's feature representation.
        """
        # generated sounds need to be re-produced for double-checking
        if not self.generated:
            if not self.alias and self.grapheme in self.ts.sounds:
                return self.grapheme
            elif self.alias and self.featureset in self.ts.features:
                return str(self.ts.features[self.featureset])
            # this can usually not happen, as we catch these errors when loading a ts!
            raise ValueError(
                'Orphaned alias {0}'.format(self.grapheme))  # pragma: no cover

        # search for best base-string
        elements = [f for f in self._features() if f not in EXCLUDE_FEATURES] + [self.type]
        base_str = self.base or '<?>'
        base_graphemes = []
        while elements:
            base = self.ts.features.get(frozenset(elements))
            if base:
                base_graphemes.append(base.grapheme)
            elements.pop(0)
        base_str = base_graphemes[-1] if base_graphemes else base_str or '<?>'
        base_vals = {
            self.ts._feature_values[elm] for elm in
            self.ts.sounds[base_str].name.split(' ')[:-1]} if \
            base_str != '<?>' else {}
        out = []
        for p in self._write_order['pre']:
            if p not in base_vals and getattr(self, p, '') in self._features():
                out.append(
                    norm(self.ts.features[self.type].get(getattr(self, p, ''), '<!>')))
        out.append(base_str)
        for p in self._write_order['post']:
            if p not in base_vals and getattr(self, p, '') in self._features():
                out.append(
                    norm(self.ts.features[self.type].get(getattr(self, p, ''), '<!>')))
        return ''.join(out)

    @property
    def name(self):
        return ' '.join([f or '' for f in self._features()] + [self.type])

    @property
    def table(self):
        """Returns the tabular representation of the sound as given in our data
        """
        tbl = []
        features = [
            f for f in self._name_order if f not in self.ts.columns[self.type]]
        # make sure to mark generated sounds
        if self.generated and self.s != self.source:
            tbl += [str(self) + ' | ' + self.source]
        else:
            tbl += [str(self)]
        for name in self.ts.columns[self.type][1:]:
            if name != 'extra' and name != 'alias':
                tbl += [getattr(self, name) or '']
            elif name == 'alias':
                tbl += ['+' if getattr(self, name) else '']
            else:
                bundle = []
                for f in features:
                    val = getattr(self, f)
                    if val:
                        bundle += ['{0}:{1}'.format(f, val)]
                tbl += [','.join(bundle)]
        return tbl


@attr.s(cmp=False)
class Marker(Symbol):
    alias = attr.ib(default=None)
    feature = attr.ib(default=None)
    value = attr.ib(default=None)
    unknown = attr.ib(default=None)

    @property
    def name(self):
        return self.grapheme

    @property
    def featureset(self):
        return frozenset([self.grapheme, self.type])


@attr.s(cmp=False, repr=False)
class Consonant(Sound):

    # features follow basic information about IPA from various sources, they
    # are potentially not yet exhaustive and should be updated at some point
    manner = attr.ib(default=None)
    place = attr.ib(default=None)
    aspiration = attr.ib(default=None)
    labialization = attr.ib(default=None)
    palatalization = attr.ib(default=None)
    preceding = attr.ib(default=None)
    velarization = attr.ib(default=None)
    duration = attr.ib(default=None)
    phonation = attr.ib(default=None)
    release = attr.ib(default=None)
    syllabicity = attr.ib(default=None)
    nasalization = attr.ib(default=None)
    glottalization = attr.ib(default=None)
    pharyngealization = attr.ib(default=None)
    ejection = attr.ib(default=None)
    voicing = attr.ib(default=None)
    breathiness = attr.ib(default=None)
    creakiness = attr.ib(default=None)
    sibilancy = attr.ib(default=None)
    laterality = attr.ib(default=None)
    laminality = attr.ib(default=None)
    articulation = attr.ib(default=None)
    raising = attr.ib(default=None)
    relative_articulation = attr.ib(default=None)

    # write order determines how consonants are written according to their
    # features, so this normalizes the order of diacritics preceding and
    # following the base part of the consonant
    _write_order = dict(
        pre=['preceding'],
        post=[
            'raising', 'relative_articulation',
            'laminality', 'creakiness', 'phonation', 'ejection', 'syllabicity', 'voicing',
            'articulation',
            'nasalization', 'palatalization', 'labialization',
            'breathiness', 'aspiration', 'glottalization', 'velarization',
            'pharyngealization', 'release', 'duration'])
    _name_order = [
        'raising', 'relative_articulation',
        'articulation', 'preceding', 'syllabicity', 'nasalization', 'palatalization',
        'labialization', 'glottalization', 'aspiration', 'velarization',
        'pharyngealization', 'duration', 'release', 'voicing', 'creakiness',
        'breathiness', 'phonation', 'laminality', 'place', 'ejection', 'laterality',
        'sibilancy', 'manner']


@attr.s(cmp=False, repr=False)
class ComplexSound(Sound):
    from_sound = attr.ib(default=None)
    to_sound = attr.ib(default=None)

    def __str__(self):
        return str(self.from_sound) + str(self.to_sound)

    @property
    def name(self):
        n1 = ' '.join(self.from_sound.name.split(' ')[:-1])
        n2 = ' '.join(self.to_sound.name.split(' ')[:-1])
        return 'from ' + n1 + ' to ' + n2 + ' ' + self.type

    @classmethod
    def from_sounds(cls, source, sound1, sound2, ts):
        return cls(
            source=source,
            grapheme=sound1.grapheme + sound2.grapheme,
            from_sound=sound1,
            to_sound=sound2,
            ts=ts,
            generated=True,
            stress=sound1.stress or sound2.stress
        )

    @property
    def table(self):
        """Overwrite the table attribute for complex sounds"""
        return [self.grapheme, self.from_sound.name, self.to_sound.name]


@attr.s(cmp=False, repr=False)
class Cluster(ComplexSound):
    """
    A cluster of two consonants whose manner is either plosive or implosive.

    Notes
    -----
    To keep the search space low and to avoid that users start defining too
    invalid sound clusters, we restrict the ```manner``` attribute of the two
    sounds to ```plosive``` and ```implosive```.
    """


@attr.s(cmp=False, repr=False)
class Vowel(Sound):
    roundedness = attr.ib(default=None)
    height = attr.ib(default=None)
    nasalization = attr.ib(default=None)
    frication = attr.ib(default=None)
    duration = attr.ib(default=None)
    voicing = attr.ib(default=None)
    breathiness = attr.ib(default=None)
    creakiness = attr.ib(default=None)
    release = attr.ib(default=None)
    syllabicity = attr.ib(default=None)
    pharyngealization = attr.ib(default=None)
    rhotacization = attr.ib(default=None)
    centrality = attr.ib(default=None)
    glottalization = attr.ib(default=None)
    velarization = attr.ib(default=None)
    relative_articulation = attr.ib(default=None)
    tone = attr.ib(default=None)
    raising = attr.ib(default=None)
    rounding = attr.ib(default=None)
    tongue_root = attr.ib(default=None)
    articulation = attr.ib(default=None)  # compare
    # https://en.wikipedia.org/wiki/Faucalized_voice

    _write_order = dict(
        pre=[],
        post=[
            'tongue_root', 'raising', 'centrality', 'rounding',
            'voicing', 'breathiness', 'creakiness',
            'syllabicity', 'nasalization', 'tone', 'articulation', 'rhotacization',
            'pharyngealization', 'glottalization', 'velarization', 'duration',
            'frication', 'relative_articulation'])
    _name_order = [
        'duration', 'rhotacization', 'pharyngealization',
        'glottalization', 'velarization', 'syllabicity',
        'relative_articulation',
        'tongue_root', 'raising', 'rounding',
        'articulation', 'nasalization', 'voicing', 'creakiness',
        'breathiness', 'roundedness', 'height', 'frication', 'centrality',
        'tone']


@attr.s(cmp=False, repr=False)
class Diphthong(ComplexSound):
    """
    A dipthong consists of two vowels.
    """


@attr.s(cmp=False, repr=False)
class Tone(Sound):
    contour = attr.ib(default=None)
    start = attr.ib(default=None)
    middle = attr.ib(default=None)
    end = attr.ib(default=None)

    _name_order = ['contour', 'start', 'middle', 'end']



from pyclts.transcriptionsystem import TranscriptionSystem  # noqa: F401
from pyclts.transcriptiondata import TranscriptionData  # noqa: F401
from pyclts.soundclasses import SoundClasses  # noqa: F401
from pyclts.api import CLTS  # noqa: F401

__version__ = '2.1.1'



"""
Prepare transcriptiondata from the transcription sources.
"""
from uritemplate import URITemplate
from clldutils.clilib import ParserError
from csvw.dsv import UnicodeWriter

try:
    from lingpy.sequence.sound_classes import token2class
    from lingpy.data import Model
    LINGPY = True
except ImportError:
    LINGPY = False
    token2class = None
    Model = None

from pyclts.soundclasses import SOUNDCLASS_SYSTEMS
from pyclts.models import is_valid_sound


def run(args):
    if not LINGPY:
        raise ParserError('lingpy must be installed to run this command!')

    def writer(*comps):
        return UnicodeWriter(args.repos.path('pkg', *comps), delimiter='\t')

    columns = ['LATEX', 'FEATURES', 'SOUND', 'IMAGE', 'COUNT', 'NOTE']
    bipa = args.repos.bipa
    for src, rows in args.repos.iter_sources(type='td'):
        args.log.info('TranscriptionData {0} ...'.format(src['NAME']))
        uritemplate = URITemplate(src['URITEMPLATE']) if src['URITEMPLATE'] else None
        out = [['BIPA_GRAPHEME', 'CLTS_NAME', 'GENERATED', 'EXPLICIT', 'GRAPHEME', 'URL'] + columns]
        graphemes = set()
        for row in rows:
            if row['GRAPHEME'] in graphemes:
                args.log.warn('skipping duplicate grapheme: {0}'.format(row['GRAPHEME']))
                continue
            graphemes.add(row['GRAPHEME'])
            if not row['BIPA']:
                bipa_sound = bipa[row['GRAPHEME']]
                explicit = ''
            else:
                bipa_sound = bipa[row['BIPA']]
                explicit = '+'
            generated = '+' if bipa_sound.generated else ''
            if is_valid_sound(bipa_sound, bipa):
                bipa_grapheme = bipa_sound.s
                bipa_name = bipa_sound.name
            else:
                bipa_grapheme, bipa_name = '<NA>', '<NA>'
            url = uritemplate.expand(**row) if uritemplate else row.get('URL', '')
            out.append(
                [bipa_grapheme, bipa_name, generated, explicit, row['GRAPHEME'],
                 url] + [
                    row.get(c, '') for c in columns])
        found = len([o for o in out if o[0] != '<NA>'])
        args.log.info('... {0} of {1} graphemes found ({2:.0f}%)'.format(
            found, len(out), found / len(out) * 100))
        with writer('transcriptiondata', '{0}.tsv'.format(src['NAME'])) as w:
            w.writerows(out)

    count = 0
    with writer('soundclasses', 'lingpy.tsv') as w:
        w.writerow(['CLTS_NAME', 'BIPA_GRAPHEME'] + SOUNDCLASS_SYSTEMS)
        for grapheme, sound in sorted(bipa.sounds.items()):
            if not sound.alias:
                w.writerow(
                    [sound.name, grapheme]
                    + [token2class(grapheme, Model(cls)) for cls in SOUNDCLASS_SYSTEMS])
                count += 1
    args.log.info('SoundClasses: {0} written to file.'.format(count))



"""

"""
from pyclts.cli_util import add_format, Table


def register(parser):
    add_format(parser)


def run(args):
    ts = args.repos.transcriptionsystem(args.system)
    features = set()
    for sound in ts.sounds.values():
        if sound.type not in ['marker', 'unknownsound']:
            for k, v in sound.featuredict.items():
                features.add((sound.type, k, v or ''))
    with Table(args, 'TYPE', 'FEATURE', 'VALUE') as table:
        table.extend(sorted(features))



import tabulate

from clldutils import markup


def add_format(parser, default='pipe'):
    parser.add_argument(
        "--format",
        default=default,
        choices=tabulate.tabulate_formats,
        help="Format of tabular output.")


class Table(markup.Table):
    def __init__(self, args, *cols):
        super().__init__(*cols)
        self._fmt = args.format

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if len(self):
            print(self.render(tablefmt=self._fmt, condensed=False, floatfmt='.2f'))



"""

"""
import collections

from pyclts.cli_util import Table, add_format


def register(parser):
    add_format(parser)
    parser.add_argument(
        '--filter',
        choices=['generated', 'unknown', 'known'],
        help='',
        default=None)
    parser.add_argument(
        'sounds',
        metavar='SOUNDS',
        nargs='+',
        help='sounds to display info for')


def run(args):
    tts = args.repos.transcriptionsystem(args.system)
    tts_sounds = [
        tts.get(sound if isinstance(sound, str) else sound.decode('utf8')) for sound in args.sounds]

    if args.filter == 'generated':
        tts_sounds = [s for s in tts_sounds if s.generated]
    elif args.filter == 'unknown':
        tts_sounds = [s for s in tts_sounds if s.type == 'unknownsound']
    elif args.filter == 'known':
        tts_sounds = [s for s in tts_sounds if not s.generated and not s.type == 'unknownsound']

    data = collections.defaultdict(list)
    ucount = 0
    for sound in tts_sounds:
        if sound.type != 'unknownsound':
            data[sound.type] += [sound.table]
        else:
            ucount += 1
            data['unknownsound'].append([str(ucount), sound.source or '', sound.grapheme])
    for cls in tts.sound_classes:
        if cls in data:
            print('# {0}\n'.format(cls))
            with Table(args, *[c.upper() for c in tts.columns[cls]]) as table:
                table.extend(data[cls])
            print('')
    if data['unknownsound']:
        print('# Unknown sounds\n')
        with Table(args, 'NUMBER', 'SOURCE', 'GRAPHEME') as table:
            table.extend(data['unknownsound'])



"""

"""
from pyclts.cli_util import add_format, Table


def register(parser):
    add_format(parser, default='simple')
    parser.add_argument(
        '--type',
        help='CLTS data type',
        choices=['sc', 'ts', 'td'],
        default='ts')


def run(args):
    with Table(args, 'id', 'description', 'refs', 'type', 'uritemplate') as table:
        for src in args.repos.meta:
            if src['TYPE'] == args.type:
                table.append(src.values())



"""
IPA charts are the most common device to visualize sound inventories.

See also https://en.wikipedia.org/wiki/International_Phonetic_Alphabet_chart
"""
import io
import copy
import pathlib
from xml.etree import ElementTree as et

import attr
from clldutils.misc import lazyproperty

__all__ = ['Segment', 'VowelTrapezoid', 'PulmonicConsonants', 'ipa_charts']


@attr.s
class Segment:
    """
    Bag of attributes controlling display of a segment in an IPA chart.
    """
    sound_bipa = attr.ib(validator=attr.validators.instance_of(str))
    sound_name = attr.ib(validator=attr.validators.instance_of(str))
    label = attr.ib(default=None)
    href = attr.ib(default=None)
    css_class = attr.ib(default=None)
    title = attr.ib(default=None)

    def __attrs_post_init__(self):
        if not self.label:
            self.label = self.sound_bipa
        if not self.title:
            self.title = self.sound_name

    @classmethod
    def from_sound(cls, sound, **kw):
        return cls(sound_bipa=str(sound), sound_name=sound.name, **kw)

    @lazyproperty
    def features(self):
        return set(s.replace('-', '') for s in self.sound_name.split())

    @property
    def link_attrib(self):
        res = {}
        if self.href:
            res['href'] = self.href
        if self.css_class:
            res['class'] = self.css_class
        return res

    @property
    def html_link_attrib(self):
        res = self.link_attrib
        res.update(title=self.title)
        return res


def html_css(id, colorspec=None):
    colorspec = colorspec or {None: ('black', 'solid 1px white')}
    res = ["""\
#{0} a {{text-decoration: none; font-size: smaller;}}
#{0} a {{color: {1};}}
#{0} a {{outline: solid 1px {2};}}
""".format(
        id,
        colorspec.get(None, ('black', None))[0],
        colorspec.get(None, (None, 'solid 1px white'))[1],
    )]
    for cls, (fill, outline) in colorspec.items():
        if cls:
            if fill:
                res.append("#{2} a.{0} text, text svg|a.{0} {{fill: {1};}}".format(cls, fill, id))
            if outline:
                res.append("#{2} a.{0} {{outline: {1};}}".format(cls, outline, id))
    return '\n'.join(res)


def svg_css(colorspec=None):
    """
    :param colorspec:
    :return:
    """
    colorspec = colorspec or {None: ('black', 'solid 1px white')}
    res = ["""\
@namespace svg url(http://www.w3.org/2000/svg);
svg|a:link, svg|a:visited {{cursor: pointer;}}
svg|a text, text svg|a {{fill: {};}}
svg|a {{outline: solid 1px {};}}
svg|a:hover, svg|a:active {{outline: dotted 1px blue;}}
""".format(
        colorspec.get(None, ('black', None))[0],
        colorspec.get(None, (None, 'white'))[1],
    )]
    for cls, (fill, outline) in colorspec.items():
        if cls:
            if fill:
                res.append("svg|a.{0} text, text svg|a.{0} {{fill: {1};}}".format(cls, fill))
            if outline:
                res.append("svg|a.{0} {{outline: {1};}}".format(cls, outline))
                # Unfortunately, there is no outline property in the SVG spec, see
                # https://stackoverflow.com/q/13387851
                # So as a fallback, we add a text-decoration:
                res.append("svg|a.{0} {{text-decoration: underline;}}".format(cls))
                res.append(
                    "svg|a.{0} {{text-decoration-style: {1};}}".format(cls, outline.split()[0]))
                res.append(
                    "svg|a.{0} {{text-decoration-color: {1};}}".format(cls, outline.split()[-1]))

    return '\n'.join(res)


class Diagram:
    """
    A visualization of a group of sounds, e.g. vowels in the "trapezoid" or consonants in a table.

    Usage:
    >>> d = Diagram()
    >>> covered = d.fill_slots(inventory)
    >>> html, css = d.render(colorspec)
    """
    __id__ = None  # HTML element id
    __fname__ = None  # Template filename
    __extend_features__ = None

    def __init__(self, id_=None):
        """
        :param id_: Pass a custom element ID to overwrite the default (e.g. to place multiple \
        diagrams of the same type on one page)
        """
        self._id = id_
        self.tree = et.parse(str(pathlib.Path(__file__).parent / self.__fname__))
        self.slots = {}
        self.exclusive = set()

    @property
    def id(self):
        return self._id or self.__id__

    def iter_slots(self):
        """
        Diagrams must provide a generator of the slots they provide as pairs (features, element),
        where `features` is a set of CLTS features (with NON<feature> specifying absence of a
        feature) and `element` is the ElementTree element where matching segments should be
        appended.
        """
        raise NotImplementedError()

    def fill_slots(self, inventory):
        """
        Assign matching segments to diagram slots.

        :param inventory: `list` of `Segment` instances.
        :return: `set` of inventory indices which have been assigned to slots.
        """
        self.slots = {}
        for features, element in self.iter_slots():
            features = set(features)
            if self.__extend_features__:
                features = features.union(self.__extend_features__)
            self.slots[frozenset(features)] = (element, [])
        covered = set()
        for i, segment in enumerate(inventory):
            features = copy.copy(segment.features)
            for ex in self.exclusive:
                if ex not in segment.features:
                    features.add('NON' + ex)
            for f in self.slots:
                if f.issubset(features):
                    covered.add(i)
                    self.slots[f][1].append(segment)
                    break
        return covered

    def format_segment(self, element, segment, is_last, is_first):
        """
        Diagrams must provide a method to format segments as ElementTree elements.
        """
        raise NotImplementedError()

    def css(self, colorspec):
        return ''

    def render(self, colorspec=None):
        for e, segments in self.slots.values():
            for i, segment in enumerate(segments, start=1):
                self.format_segment(e, segment, i == len(segments), i == 0)

        self.tree.getroot().attrib['id'] = self.id
        o = io.BytesIO()
        self.tree.write(o)
        return o.getvalue().decode('utf8'), self.css(colorspec)


class PulmonicConsonants(Diagram):
    __id__ = 'pulmonic-consonants'
    __fname__ = 'consonants.html'
    __extend_features__ = frozenset({'consonant'})

    def iter_slots(self):
        for e in self.tree.findall('.//td'):
            if 'class' in e.attrib:
                for attrs in e.attrib['class'].split():
                    attrs = attrs.split('-')
                    for att in attrs:
                        if att.startswith('NON'):
                            self.exclusive.add(att[3:])
                    yield attrs, e

    def format_segment(self, e, segment, is_last, is_first):
        ee = et.SubElement(e, 'a', attrib=segment.html_link_attrib)
        ee.text = segment.label
        if not is_last:
            ee.tail = '\xa0'

    def css(self, colorspec):
        return html_css(self.id, colorspec)


class VowelTrapezoid(Diagram):
    __id__ = 'vowel-trapezoid'
    __fname__ = 'vowels.svg'
    __extend_features__ = frozenset({'vowel'})

    def iter_slots(self):
        ns = {'svg': "http://www.w3.org/2000/svg"}
        et.register_namespace('', ns['svg'])
        for e in self.tree.findall('.//svg:text', ns):
            if 'id' in e.attrib:
                yield e.attrib['id'].split('-'), e

    def format_segment(self, e, segment, is_last, is_first):
        ee = et.SubElement(e, '{http://www.w3.org/2000/svg}a', attrib=segment.link_attrib)
        title = et.SubElement(ee, '{http://www.w3.org/2000/svg}title')
        title.text = segment.title
        ee.text = segment.label
        if not is_last:
            ee.tail = ' '

    def css(self, colorspec):
        return """\
#{0} {{height: 300px; width: 100%; min-width: 800px;}}
#{0} .label {{font-size: 150%;}}
#{0} .glyph {{font-size: 170%;}}
""".format(self.id)

    def render(self, colorspec=None):
        r = self.tree.getroot()
        del r.attrib['width']
        del r.attrib['height']
        style = et.SubElement(r, '{http://www.w3.org/2000/svg}style')
        style.text = svg_css(colorspec)
        res, css = Diagram.render(self)
        return '<figure>{}<figcaption>Vowels</figcaption></figure>'.format(
            res.replace('#666666', '#dddddd')), css


def ipa_charts(inventory, colorspec=None):
    """
    Slots matching segments into a set of predefined diagrams.

    :param inventory:
    :param colorspec:
    :return: A pair (html, covered)
    """
    css, html, covered = [], [], set()

    for diagram in [
        PulmonicConsonants(),
        VowelTrapezoid(),
    ]:
        covered = covered.union(diagram.fill_slots(inventory))
        html_, css_ = diagram.render(colorspec)
        html.append(html_)
        css.append(css_)

    return """\
<html>
<head>
<style>
body {{font-family: sans-serif}}
table caption {{text-align: left;}}
figure {{display: table; margin-left: 0px;}}
figcaption {{display: table-caption; caption-side: top; font-size: 120%;}}
{}
</style>
</head>
<body>
{}
</body>
</html>""".format(
        '\n'.join(css),
        '\n'.join('<div>{}</div>'.format(t) for t in html)), covered



"""Auxiliary functions for pyclts."""

import unicodedata
from collections import defaultdict
from pathlib import Path

from csvw.dsv import reader

__all__ = ['EMPTY', 'UNKNOWN', 'norm', 'nfd', 'TranscriptionBase']

EMPTY = "◌"
UNKNOWN = "�"


class TranscriptionBase(object):
    __type__ = None

    def __init__(self, path, system=None):
        self.path = Path(path)
        self.system = system

    @property
    def id(self):
        return self.path.stem

    def resolve_sound(self, sound):
        raise NotImplementedError  # pragma: no cover

    def __getitem__(self, sound):
        """Return a Sound instance matching the specification."""
        return self.resolve_sound(sound)

    def get(self, sound, default=None):
        try:
            res = self[sound]
            if getattr(res, 'type', None) == 'unknownsound' and default:
                return default
            return res
        except KeyError:
            return default

    def __call__(self, sounds, default="0"):
        return [self.get(x, default=default) for x in sounds.split()]

    def translate(self, string, target_system):
        return ' '.join('{0}'.format(
            target_system.get(self[s].name or '?', '?')) for s in string.split())


def norm(string):
    return string.replace(EMPTY, "")


def nfd(string):
    return unicodedata.normalize("NFD", string)


def itertable(table):
    """Auxiliary function for iterating over a data table."""
    for item in table:
        res = {
            k.lower(): nfd(v) if isinstance(v, str) else v for k, v in item.items()}
        for extra in res.pop('extra', []):
            k, _, v = extra.partition(':')
            res[k.strip()] = v.strip()
        yield res


def read_data(fname, grapheme_col, *cols):
    grapheme_map, data, sounds, names = {}, defaultdict(list), [], []

    for row in reader(fname, delimiter='\t', dicts=True):
        grapheme_map[row[grapheme_col]] = row['BIPA_GRAPHEME']
        grapheme = {"grapheme": row[grapheme_col]}
        for col in cols:
            grapheme[col.lower()] = row[col]
        data[row['BIPA_GRAPHEME']].append(grapheme)
        data[row['CLTS_NAME']].append(grapheme)
        sounds.append(row['BIPA_GRAPHEME'])
        names.append(row['CLTS_NAME'])

    return grapheme_map, data, sounds, names



"""
Dump the data in the CLTS collection for convenient reuse.
"""
import json
import zipfile
import collections

import attr
from csvw.dsv import UnicodeWriter
from clldutils.clilib import PathType

from pyclts.models import is_valid_sound


def register(parser):
    parser.add_argument(
        "--destination",
        default=None,
        type=PathType(type='file', must_exist=False),
        help="Name of the file to store data in compressed form."
    )


@attr.s
class Grapheme(object):
    GRAPHEME = attr.ib()
    NAME = attr.ib()
    EXPLICIT = attr.ib()
    ALIAS = attr.ib()
    DATASET = attr.ib()
    FREQUENCY = attr.ib(default=0)
    URL = attr.ib(default='')
    FEATURES = attr.ib(default='')
    IMAGE = attr.ib(default='')
    SOUND = attr.ib(default='')
    NOTE = attr.ib(default='')


def run(args):
    args.destination = args.destination or args.repos.path('data', 'clts.zip')

    def writer(*comps):
        return UnicodeWriter(args.repos.path('data', *comps), delimiter='\t')

    sounds = collections.defaultdict(dict)
    data = []
    clts_dump = collections.OrderedDict()
    bipa = args.repos.bipa
    # start from assembling bipa-sounds
    args.log.info('adding bipa data')
    for grapheme, sound in sorted(
            bipa.sounds.items(), key=lambda p: p[1].alias if p[1].alias else False):
        if sound.type not in ['marker']:
            if sound.alias:
                assert sound.name in sounds
                sounds[sound.name]['aliases'].add(grapheme)
            else:
                assert sound.name not in sounds
                sounds[sound.name] = {
                    'grapheme': grapheme,
                    'unicode': sound.uname or '',
                    'generated': '',
                    'note': sound.note or '',
                    'type': sound.type,
                    'aliases': set(),
                    'normalized': '+' if sound.normalized else ''
                }
            data.append(Grapheme(
                grapheme,
                sound.name,
                '+',
                '',
                'bipa',
                '0',
                '',
                '',
                '',
                '',
                sound.note or ''))
            if grapheme not in clts_dump:
                clts_dump[grapheme] = [str(sound), sound.name]

    # add sounds systematically by their alias
    args.log.info('adding transcription data')
    for td in args.repos.iter_transcriptiondata():
        for name in td.names:
            bipa_sound = bipa[name]
            # check for consistency of mapping here
            if not is_valid_sound(bipa_sound, bipa):
                continue

            sound = sounds.get(name)
            if not sound:
                sound = sounds[name] = {
                    'grapheme': bipa_sound.s,
                    'aliases': {bipa_sound.s},
                    'generated': '+',
                    'unicode': bipa_sound.uname or '',
                    'note': '',
                    'type': bipa_sound.type,
                    'alias': '+' if bipa_sound.alias else '',
                    'normalized': '+' if bipa_sound.normalized else ''
                }

            for item in td.data[name]:
                sound['aliases'].add(item['grapheme'])
                # add the values here
                data.append(Grapheme(
                    item['grapheme'],
                    name,
                    item['explicit'],
                    '',  # sounds[name]['alias'],
                    td.id,
                    item.get('frequency', ''),
                    item.get('url', ''),
                    item.get('features', ''),
                    item.get('image', ''),
                    item.get('sound', ''),
                ))
                if item['grapheme'] not in clts_dump:
                    clts_dump[item['grapheme']] = [sound['grapheme'], name]

    # sound classes have a generative component, so we need to treat them
    # separately
    args.log.info('adding sound classes')
    for sc in args.repos.iter_soundclass():
        for name in sounds:
            try:
                grapheme = sc[name]
                data.append(Grapheme(
                    grapheme,
                    name,
                    '+' if name in sc.data else '',
                    '',
                    sc.id,
                ))
            except KeyError:  # pragma: no cover
                args.log.debug(name, sounds[name]['grapheme'])

    # last run, check again for each of the remaining transcription systems,
    # whether we can translate the sound
    args.log.info('adding remaining transcriptin systems')
    for ts in args.repos.iter_transcriptionsystem(exclude=['bipa']):
        for name in sounds:
            try:
                ts_sound = ts[name]
                if is_valid_sound(ts_sound, ts):
                    sounds[name]['aliases'].add(ts_sound.s)
                    data.append(Grapheme(
                        ts_sound.s,
                        name,
                        '' if sounds[name]['generated'] else '+',
                        '',  # sounds[name]['alias'],
                        ts.id,
                    ))
                    if ts_sound.s not in clts_dump:
                        clts_dump[ts_sound.s] = [sounds[name]['grapheme'], name]
            except ValueError:
                pass
            except TypeError:  # pragma: no cover
                args.log.debug('{0}: {1}'.format(ts.id, name))

    args.log.info('writing data to file')
    with writer('sounds.tsv') as w:
        w.writerow(['NAME', 'TYPE', 'GRAPHEME', 'UNICODE', 'GENERATED', 'NOTE'])
        for k, v in sorted(sounds.items(), reverse=True):
            w.writerow([k, v['type'], v['grapheme'], v['unicode'], v['generated'], v['note']])

    with writer('graphemes.tsv') as w:
        w.writerow([f.name for f in attr.fields(Grapheme)])
        for row in data:
            w.writerow(attr.astuple(row))

    with zipfile.ZipFile(
        str(args.destination),
        mode='w',
        compression=zipfile.ZIP_DEFLATED
    ) as myzip:
        myzip.writestr('clts.json', json.dumps(clts_dump))



"""
Check repository data for consistency
"""
import argparse
import pathlib

from csvw.dsv import reader

import pyclts


def register(parser):
    parser.add_argument(
        '--test',
        action='store_true',
        default=False,
        help=argparse.SUPPRESS,
    )


def run(args):
    clts = args.repos

    if not args.test:  # pragma: no cover
        test_transcriptiondata(
            clts.soundclass('sca'),
            clts.soundclass('dolgo'),
            clts.soundclass('asjp'),
            clts.transcriptiondata('phoible'),
            clts.transcriptionsystem('bipa'))
        test_transcription_system_consistency(
            *[clts.transcriptionsystem(key) for key in ['bipa', 'asjpcode', 'gld']])
    test_sounds(clts.bipa, args.log)
    test_clicks(clts.bipa)


def test_transcriptiondata(sca, dolgo, asjpd, phoible, bipa):  # pragma: no cover
    seq = 'tʰ ɔ x ˈth ə r A ˈI ʲ'
    seq2 = 'th o ?/x a'
    seq3 = 'th o ?/ a'
    seq4 = 'ǃŋ i b ǃ'

    assert dolgo(seq) == list('TVKTVR000')
    assert sca(seq2)[2] == 'G'
    assert asjpd(seq2)[2] == 'x'
    assert sca(seq3)[2] == '0'

    # these tests need to be adjusted once lingpy accepts click sounds
    assert sca(seq4)[0] == '0'
    assert asjpd(seq4)[0] == '0'
    assert sca(seq4)[3] == '!'
    assert asjpd(seq4)[3] == '!'

    # test data from sound name
    assert sca.resolve_sound(bipa['ʰb']) == 'P'
    assert sca.resolve_sound(bipa['ae']) == 'A'
    assert sca.resolve_sound(bipa['tk']) == 'T'

    assert phoible.resolve_sound('m') == 'm'
    try:
        phoible.resolve_sound(bipa['tk'])
        raise ValueError()
    except KeyError:
        pass


def test_transcription_system_consistency(bipa, asjp, gld):  # pragma: no cover
    # bipa should always be able to be translated to
    for sound in asjp:
        if sound not in bipa:
            assert '<?>' not in str(bipa[asjp[sound].name])
    for sound in gld:
        if sound not in bipa:
            assert '<?>' not in str(bipa[gld[sound].name])
    for sound in bipa:
        if bipa[sound].type != 'unknownsound' and not bipa[sound].alias:
            if sound != str(bipa[sound]):
                raise ValueError
        elif bipa[sound].type == 'unknownsound':
            raise ValueError
    for sound in gld:
        if gld[sound].type != 'unknownsound' and not gld[sound].alias:
            if sound != str(gld[sound]):
                raise ValueError
        elif gld[sound].type == 'unknownsound':
            raise ValueError
    for sound in asjp:
        if asjp[sound].type != 'unknownsound' and not asjp[sound].alias:
            if sound != str(asjp[sound]):
                raise ValueError
        elif asjp[sound].type == 'unknownsound':
            raise ValueError

    # important test for alias
    assert str(bipa['d̤ʷ']) == str(bipa['dʷʱ']) == str(bipa['dʱʷ'])


def read_tests(name):
    return reader(pathlib.Path(pyclts.__file__).parent / 'data' / name, delimiter='\t', dicts=True)


def test_sounds(bipa, log):
    for test in read_tests('test_data.tsv'):
        del test['bipa']
        if None in test:
            del test[None]
        try:
            _test_sounds(bipa, **{k.replace('-', '_'): v for k, v in test.items()})
        except AssertionError as e:
            log.warning('{0}\t{1}'.format(test['source'], e))


def test_clicks(bipa):
    for test in read_tests('clicks.tsv'):
        _test_clicks(bipa, test['GRAPHEME'], test['MANNER'])


def _test_clicks(bipa, grapheme, gtype):
    if gtype == 'stop-cluster':
        assert bipa[grapheme].type == 'cluster'


def _test_sounds(bipa, **kw):
    """Test on a large pre-assembled dataset whether everything is consistent"""
    kw = argparse.Namespace(**kw)

    sound = bipa[kw.source]
    if sound.type not in ['unknownsound', 'marker']:
        if kw.nfd_normalized == '+':
            assert bipa[kw.source] != sound.source, "Sound does not resolve to itself"
        if kw.clts_normalized == "+":
            assert sound.normalized, "Sound not normalized"
        if kw.aliased == '+':
            assert sound.alias, "Sound not an alias"
        if kw.generated:
            assert sound.generated, "Sound not generated"
        if kw.stressed:
            assert sound.stress, "Sound not stressed"
        assert sound.name == kw.name
        assert sound.codepoints == kw.codepoints, "Not matching codepoints"



from clldutils.apilib import API
from clldutils.misc import lazyproperty
from csvw.dsv import iterrows as reader
from cldfcatalog import Config

from pyclts import TranscriptionData, TranscriptionSystem, SoundClasses
from pyclts.soundclasses import SOUNDCLASS_SYSTEMS


class CLTS(API):
    def __init__(self, repos=None):
        if repos is None:
            repos = Config.from_file().get_clone('clts')  # pragma: no cover
        super().__init__(repos)
        self.pkg_dir = self.repos / 'pkg'
        self.transcriptionsystems_dir = self.pkg_dir / 'transcriptionsystems'
        self.transcriptiondata_dir = self.pkg_dir / 'transcriptiondata'
        self.soundclasses_dir = self.pkg_dir / 'soundclasses'

    @lazyproperty
    def bipa(self):
        return self.transcriptionsystem('bipa')

    @lazyproperty
    def meta(self):
        return list(reader(self.repos / 'sources' / 'index.tsv', dicts=True, delimiter='\t'))

    def get_meta(self, obj):
        for src in self.meta:
            if obj.__type__ == src['TYPE'] and obj.id == src['NAME']:
                return src

    def iter_sources(self, type=None):
        for src in self.meta:
            if (type is None) or (type == src['TYPE']):
                graphemesp = self.repos / 'sources' / src['NAME'] / 'graphemes.tsv'
                if graphemesp.exists():
                    yield src, list(reader(graphemesp, dicts=True, delimiter='\t'))

    def iter_transcriptiondata(self):
        for td in sorted(self.transcriptiondata_dir.iterdir(), key=lambda p: p.name):
            yield TranscriptionData(td, self.bipa)

    def iter_soundclass(self):
        for sc in SOUNDCLASS_SYSTEMS:
            yield SoundClasses(self.soundclasses_dir / 'lingpy.tsv', self.bipa, sc)

    def iter_transcriptionsystem(self, include_private=False, exclude=None):
        exclude = exclude or []
        for ts in sorted(self.transcriptionsystems_dir.iterdir(), key=lambda p: p.name):
            if ts.is_dir():
                if (not ts.name.startswith('_')) or include_private:
                    if ts.name not in exclude:
                        yield TranscriptionSystem(
                            ts,
                            self.transcriptionsystems_dir / 'transcription-system-metadata.json',
                            self.transcriptionsystems_dir / 'features.json',
                        )

    @lazyproperty
    def transcriptionsystem_dict(self):
        return {ts.id: ts for ts in self.iter_transcriptionsystem()}

    def transcriptionsystem(self, key):
        if key in self.transcriptionsystem_dict:
            return self.transcriptionsystem_dict[key]
        return TranscriptionSystem(
            key,
            self.transcriptionsystems_dir / 'transcription-system-metadata.json',
            self.transcriptionsystems_dir / 'features.json',
        )

    @lazyproperty
    def transcriptiondata_dict(self):
        return {ts.id: ts for ts in self.iter_transcriptiondata()}

    def transcriptiondata(self, key):
        if key in self.transcriptiondata_dict:
            return self.transcriptiondata_dict[key]
        return TranscriptionData(key, self.bipa)

    @lazyproperty
    def soundclasses_dict(self):
        return {ts.id: ts for ts in self.iter_soundclass()}

    def soundclass(self, key):
        return self.soundclasses_dict[key]



from pyclts.util import read_data, TranscriptionBase
from pyclts.transcriptionsystem import Sound


class TranscriptionData(TranscriptionBase):
    """
    Class for handling transcription data.
    """
    __type__ = 'td'

    def __init__(self, path, system):
        super().__init__(path, system)
        self.grapheme_map, self.data, self.sounds, self.names = read_data(
            self.path,
            'GRAPHEME',
            'URL',
            'BIPA_GRAPHEME',
            'GENERATED',
            'URL',
            'LATEX',
            'FEATURES',
            'SOUND',
            'IMAGE',
            'COUNT',
            'NOTE',
            'EXPLICIT'
        )

    def resolve_sound(self, sound):
        """Function tries to identify a sound in the data.

        Notes
        -----
        The function tries to resolve sounds to take a sound with less complex
        features in order to yield the next approximate sound class, if the
        transcription data are sound classes.
        """
        if not isinstance(sound, Sound):
            sound = self.system[sound]
        if sound.name in self.data:
            return '//'.join([x['grapheme'] for x in self.data[sound.name]])
        raise KeyError(":td:resolve_sound: No sound could be found.")

    def resolve_grapheme(self, grapheme):
        return self.system[self.grapheme_map[grapheme]]
